import asyncio
import logging
import shlex
import subprocess
import uuid
from asyncio.subprocess import Process
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Callable, Optional, Union

from .helpers import is_test

process_pool = ProcessPoolExecutor()
thread_pool = ThreadPoolExecutor(thread_name_prefix='run.py thread_pool')
# NOTE is used in rosys.test.Runtime to advance time slower until computation is done
running_cpu_bound_processes: list[int] = []
running_sh_processes: list[Process] = []
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: any):
    loop = asyncio.get_running_loop()
    try:
        return await loop.run_in_executor(thread_pool, callback, *args)
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    except asyncio.exceptions.CancelledError:
        pass


async def cpu_bound(callback: Callable, *args: any):
    with cpu():
        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(process_pool, callback, *args)
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise
        except asyncio.exceptions.CancelledError:
            pass


@contextmanager
def cpu():
    id = str(uuid.uuid4())
    running_cpu_bound_processes.append(id)
    try:
        yield
    finally:
        running_cpu_bound_processes.remove(id)


async def sh(command: Union[list[str], str], timeout: Optional[float] = 1, shell=False) -> str:
    '''executes a shell command
    command: a sequence of program arguments as subprocess.Popen requires or full string
    shell: weather a sub shell should be launched (default is False, for speed, use True if you need file globbing or other features);
    returns: stdout
    '''
    def popen() -> str:
        if shell:  # convert to string
            cmd = (' '.join(command)) if isinstance(command, list) else command
        else:  # convert to list
            cmd = shlex.split(command) if isinstance(command, str) else command
        if timeout is not None:
            if shell:
                cmd = f'timeout {timeout} {cmd}'
            else:
                cmd = ['timeout', str(timeout)] + cmd
        #log.info(f'running sh: "{cmd}"')
        with subprocess.Popen(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            shell=shell,
        ) as proc:
            stdout, *_ = proc.communicate()
            return stdout.decode('utf-8')
    return await io_bound(popen)


def tear_down() -> None:
    log.info('teardown thread_pool')
    thread_pool.shutdown(wait=False, cancel_futures=True)
    [p.kill() for p in running_sh_processes]
    if not is_test:
        log.info('teardown process_pool')
        [p.kill() for p in process_pool._processes.values()]
        process_pool.shutdown(wait=True, cancel_futures=True)
    log.info('teardown complete')



#!/usr/bin/env python3
from nicegui import ui
import pylab as pl
from rosys.actors.pathplanning import generate_experiment, plot_tools as pt

robot_renderer, pose, goal, obstacle_map, small_obstacle_map, backward_to_goal = generate_experiment(5.0)

with ui.plot():
    pt.show_obstacle_map(obstacle_map)
    pl.gca().invert_yaxis()
    pl.autoscale(False)
    pt.plot_robot(robot_renderer, pose, 'C0')
    pt.plot_robot(robot_renderer, goal, 'C1')

ui.run()



from __future__ import annotations

import io
import os
from datetime import datetime
from glob import glob
from pathlib import Path
from typing import Optional

import humanize
import rosys
from PIL import Image, ImageDraw, ImageFont
from rosys.actors import Actor

from ..world import World

rosys_dir = os.path.dirname(os.path.dirname(__file__))
font = f'{rosys_dir}/RobotoMono-Medium.ttf'
image_font = ImageFont.truetype(font, 12)
big_cover_font = ImageFont.truetype(font, 100)
small_cover_font = ImageFont.truetype(font, 60)


class TimelapseRecorder(Actor):
    interval: float = 1
    world: World
    storage_path: str = os.path.expanduser('~/.rosys/timelapse')

    def __init__(self) -> None:
        super().__init__()
        os.makedirs(self.storage_path + '/videos', exist_ok=True)

    async def save(self, image: rosys.Image) -> None:
        await rosys.run.cpu_bound(save_image, image, self.storage_path)

    async def compress_video(self) -> None:
        jpgs = sorted(glob(f'{self.storage_path}/*.jpg'))
        if len(jpgs) < 2:
            return
        start_unix = int(Path(jpgs[0]).stem[:-4])
        start = datetime.fromtimestamp(start_unix)
        end = datetime.fromtimestamp(int(Path(jpgs[-1]).stem[:-4]))
        self.log.info(f'creating video from {start} to {end}')
        duration = humanize.naturaldelta(end-start)
        await self.create_info(start.strftime('%d.%m.%Y %H:%M:%S'), duration, time=start_unix - 1)
        id = start.strftime('%Y%m%d_%H-%M-%S_' + duration.replace(' ', '_'))
        target_dir = self.storage_path + '/' + id
        os.makedirs(target_dir, exist_ok=True)
        os.makedirs(self.storage_path + '/videos', exist_ok=True)
        self.log.info(await rosys.run.sh(f'mv {self.storage_path}/*.jpg {target_dir}', shell=True))
        cmd = f'nice -n 19 ffmpeg -hide_banner -r 10 -pattern_type glob -i "{target_dir}/*.jpg" -s 1600x1200 -vcodec libx264 -crf 18 -preset slow -pix_fmt yuv420p -y {target_dir}/{id}.mp4; mv {target_dir}/*mp4 {self.storage_path}/videos; rm -r {target_dir};'
        self.log.info(f'starting {cmd}')
        rosys.task_logger.create_task(rosys.run.sh(cmd, timeout=None, shell=True), name='timelapse ffmpeg')

    def clear_jpegs(self) -> None:
        files = glob(f'{self.storage_path}/**/*.jpg', recursive=True)
        for f in files:
            os.remove(f)

    async def create_info(self, title: str, subtitle: str, frames: int = 20, time: Optional[int] = None) -> None:
        prefix = f'{self.storage_path}/{int(self.world.time if time is None else time)}'
        await rosys.run.cpu_bound(save_info, title, subtitle, prefix, frames)


def save_image(image: rosys.world.Image, path: str):  # static to run cpu bound
    img = Image.open(io.BytesIO(image.data))
    draw = ImageDraw.Draw(img)
    x = img.width - 300
    y = img.height - 18
    text = f'{datetime.fromtimestamp(image.time).strftime("%Y-%m-%d %H:%M:%S")}, cam {image.camera_id}'
    # shadow
    draw.text((x - 1, y - 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x + 1, y - 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x - 1, y + 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x + 1, y + 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x, y), text, font=image_font, fill=(255, 255, 255))
    filepath = path + f'/{int(image.time)}0000.jpg'
    img.save(filepath, 'JPEG')


def save_info(title: str, subtitle: str, prefix: str, frames: int):
    for i in range(frames):
        img = Image.new('RGB', (1600, 1200), 'black')
        draw = ImageDraw.Draw(img)
        font = big_cover_font if len(subtitle) < 25 and len(title) < 25 else small_cover_font
        w, h = draw.textsize(title, font=font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 - 200), title, font=font, fill='white')
        w, h = draw.textsize(subtitle, font=font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 + 100), subtitle, font=font, fill='white')
        filepath = f'{prefix}{i:04}.jpg'
        img.save(filepath)



import inspect
from typing import Awaitable, Callable, Optional, Union

from nicegui.ui import Ui
from rosys.hardware import AppButton, AppControls

from .. import Runtime
from .settings import Settings as settings


class AutomationControls:
    ui: Ui
    runtime: Runtime

    def __init__(self, *,
                 default_automation: Optional[Awaitable] = None,
                 can_start: Optional[Callable[[], Union[bool, Awaitable[bool]]]] = None,
                 app_controls: AppControls = None,
                 ) -> None:

        async def start():
            if inspect.iscoroutinefunction(can_start):
                if not await can_start():
                    return
            elif can_start is not None and not can_start():
                return
            self.runtime.automator.start(default_automation())

        def pause():
            self.runtime.automator.pause(because='pause button was pressed')

        def resume():
            self.runtime.automator.resume()

        def stop():
            self.runtime.automator.stop(because='stop button was pressed')

        play_button = self.ui.button(on_click=start).props('icon=play_arrow unelevated').tooltip('start automation')
        pause_button = self.ui.button(on_click=pause).props('icon=pause outline').tooltip('stop automation')
        resume_button = self.ui.button(on_click=resume).props('icon=play_arrow outline').tooltip('resume automation')
        stop_button = self.ui.button(on_click=stop).props('icon=stop outline').tooltip('stop automation')

        if app_controls is not None:
            app_controls.main_buttons['play'] = AppButton('play_arrow', released=start)
            app_controls.main_buttons['pause'] = AppButton('pause', released=pause)
            app_controls.main_buttons['resume'] = AppButton('play_arrow', released=resume)
            app_controls.main_buttons['stop'] = AppButton('stop', released=stop)

        async def refresh():
            play_button.visible = self.runtime.automator.is_stopped
            pause_button.visible = self.runtime.automator.is_running
            resume_button.visible = self.runtime.automator.is_paused
            play_button.view.disable = default_automation is None or not self.runtime.automator.enabled
            stop_button.view.disable = self.runtime.automator.is_stopped
            if app_controls is not None:
                before = ' '.join(str(b) for b in app_controls.main_buttons.values())
                app_controls.main_buttons['play'].visible = play_button.visible
                app_controls.main_buttons['pause'].visible = pause_button.visible
                app_controls.main_buttons['resume'].visible = resume_button.visible
                app_controls.main_buttons['play'].state = 'disabled' if play_button.view.disable else 'enabled'
                app_controls.main_buttons['stop'].state = 'disabled' if stop_button.view.disable else 'enabled'
                after = ' '.join(str(b) for b in app_controls.main_buttons.values())
                if after != before:
                    await app_controls.sync()
        self.ui.timer(settings.update_interval, refresh)
